from __future__ import absolute_import, unicode_literals

from celery import shared_task

from .loaders import load_lau as _load_lau, load_nuts as _load_nuts


@shared_task
def load_lau(code):
    return _load_lau(code)


@shared_task
def load_nuts(code):
    return _load_nuts(code)



from django.db import models
from django.utils.functional import cached_property


class NUTS(models.Model):
    code = models.CharField(max_length=6, primary_key=True)
    name = models.CharField(max_length=250)
    level = models.PositiveSmallIntegerField()

    class Meta:
        ordering = ('code',)
        verbose_name = 'NUTS'
        verbose_name_plural = 'NUTS'

    def __str__(self):
        return self.name

    @cached_property
    def laus(self):
        return LAU.objects.filter(nuts__code__startswith=self.code)


class LAU(models.Model):
    code = models.CharField(max_length=15, unique=True)
    name = models.CharField(max_length=250)
    local_name = models.CharField(max_length=250)
    nuts = models.ForeignKey(NUTS, related_name='+')

    class Meta:
        ordering = ('code',)
        verbose_name = 'LAU'
        verbose_name_plural = 'LAU'

    def __str__(self):
        return self.name

    @cached_property
    def nuts0(self):
        return NUTS.objects.get(code=self.nuts_id[:2])

    @cached_property
    def nuts1(self):
        return NUTS.objects.get(code=self.nuts_id[:3])

    @cached_property
    def nuts2(self):
        return NUTS.objects.get(code=self.nuts_id[:4])

    @cached_property
    def nuts3(self):
        if len(self.nuts_id) == 5:
            return self.nuts
        else:
            return NUTS.objects.get(code=self.nuts_id[:5])

    @cached_property
    def nuts4(self):
        return self.nuts if len(self.nuts_id) == 6 else None



import logging
import os
from csv import DictReader
from io import StringIO

from .models import LAU, NUTS

try:
    # python3
    from urllib.request import urlopen
except ImportError:
    # python2
    from urllib import urlopen


logger = logging.getLogger(__name__)

NUTS_URL = os.environ.get(
    'NUTS_URL',
    'http://ec.europa.eu/eurostat/ramon/nomenclatures/index.cfm'
    '?TargetUrl=ACT_OTH_CLS_DLD&StrNom=NUTS_2016L&StrFormat=CSV&StrLanguageCode=EN&IntKey=&IntLevel=&bExport='
)

LAU_URL = os.environ.get(
    'LAU_URL',
    'http://ec.europa.eu/eurostat/documents/345175/501971/EU-28_LAU_2017_NUTS_2016.xlsx',
)


def load_nuts(*country_codes):
    with urlopen(NUTS_URL) as response:
        content = response.read().replace(b'\r', b'\n').decode('utf-8')
    nuts = {n.code: n for n in NUTS.objects.iterator()}
    created, updated = 0, 0
    for record in sorted(DictReader(StringIO(content), delimiter=';'), key=lambda r: r['NUTS-Code']):
        code = record['NUTS-Code']
        level = len(code) - 2
        name = record['Description']

        # skip records not matching base_code and records containing NUTS in the name
        if country_codes and code[:2] not in country_codes or 'NUTS' in name:
            continue

        if code in nuts:
            if nuts[code].name != name:
                nuts[code].name = name
                nuts[code].save()
                updated += 1
        else:
            nuts[code] = NUTS.objects.create(code=code, name=name, level=level)
            created += 1
    logger.info('Created %d NUTS, updated %d', created, updated)


def load_lau(*country_codes):
    data = get_remote_data(LAU_URL)
    if not country_codes:
        country_codes = (key for key in data.keys() if len(key) == 2)
    for country_code in country_codes:
        sheet = data[country_code]
        nuts = {n.code: n for n in NUTS.objects.filter(code__startswith=country_code).iterator()}
        laus = {l.code: l for l in LAU.objects.filter(code__startswith=country_code).iterator()}
        created, updated = 0, 0
        for row in sheet[1:]:
            # handle exceptions in data
            try:
                nuts_code, lau_code, local_name, name = row[:4]
                if not local_name:
                    local_name = name
                if not name:
                    name = local_name
            except ValueError:
                nuts_code, lau_code, local_name = row[:3]
                name = local_name
            if not all((nuts_code, lau_code, local_name, name)) or len(row[:4]) < 3:
                logger.warning('Incomplete row in %s: %s', country_code, row[:4])
                continue
            if nuts_code[-3:] == 'ZZZ':
                continue
            if nuts_code not in nuts:
                logger.warning('Unknown NUTS-3 code: %s', nuts_code)
                continue
            # store data
            code = '%s%s' % (country_code, lau_code)
            if code in laus:
                lau = laus[code]
                if not lau.nuts_id.startswith(nuts_code) or lau.name != name or lau.local_name != local_name:
                    if not lau.nuts_id.startswith(nuts_code):
                        lau.nuts2 = nuts[nuts_code]
                    lau.name = name
                    lau.local_name = local_name
                    lau.save()
                    updated += 1
            else:
                laus[code] = LAU.objects.create(
                    code=code,
                    name=name,
                    local_name=local_name,
                    nuts=nuts[nuts_code],
                )
                created += 1
        logger.info('Created %d LAU, updated %d in %s', created, updated, country_code)


def get_remote_data(url, suffix='.xlsx'):
    from pyexcel_xls import get_data
    from tempfile import NamedTemporaryFile

    with NamedTemporaryFile(suffix=suffix) as f:
        with urlopen(url) as response:
            f.write(response.read())
            f.flush()
        return get_data(f.name)



from django.test import TestCase


class SimpleTest(TestCase):
    def test_load_cz(self):
        from django_nuts.loaders import load_lau, load_nuts
        from django_nuts.cz_nuts4 import load_cz_nuts4
        from django_nuts.models import LAU, NUTS
        load_nuts('CZ', 'SK')
        load_lau('CZ', 'SK')
        load_cz_nuts4()
        self.assertTrue(NUTS.objects.count() > 20)
        self.assertTrue(LAU.objects.count() > 5000)
        load_nuts()
        load_lau()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-02-12 00:55
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LAU',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=15, unique=True)),
                ('name', models.CharField(max_length=250)),
                ('local_name', models.CharField(max_length=250)),
            ],
            options={
                'verbose_name': 'LAU',
                'verbose_name_plural': 'LAU',
                'ordering': ('code',),
            },
        ),
        migrations.CreateModel(
            name='NUTS',
            fields=[
                ('code', models.CharField(max_length=6, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=250)),
                ('level', models.PositiveSmallIntegerField()),
            ],
            options={
                'verbose_name': 'NUTS',
                'verbose_name_plural': 'NUTS',
                'ordering': ('code',),
            },
        ),
        migrations.AddField(
            model_name='lau',
            name='nuts',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='django_nuts.NUTS'),
        ),
    ]



from django.contrib import admin

from .models import LAU, NUTS


class LAUAdmin(admin.ModelAdmin):
    list_display = ('code', 'name', 'local_name', 'nuts1', 'nuts2', 'nuts3', 'nuts4')
    search_fields = ('nuts__code', 'code', 'name', 'local_name')


class NUTSAdmin(admin.ModelAdmin):
    list_display = ('code', 'name', 'level')
    list_filter = ('level',)
    search_fields = ('code', 'name')


admin.site.register(LAU, LAUAdmin)
admin.site.register(NUTS, NUTSAdmin)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    long_description = readme.read()

setup(
    name='django-nuts',
    version='1.1.1',
    description='Django application providing database of European NUTS and LAU',
    long_description=long_description,
    author='Jakub Dorňák',
    author_email='jakub.dornak@misli.cz',
    license='BSD',
    url='https://github.com/misli/django-nuts',
    packages=find_packages(),
    install_requires=[
        'pyexcel-xls'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
)



import logging
import os

from .loaders import get_remote_data
from .models import LAU, NUTS

logger = logging.getLogger(__name__)

CZ_NUTS4_URL = os.environ.get(
    'CZ_NUTS4_URL',
    'https://www.czso.cz/documents/10180/23208674/struktura_uzemi_cr_1_1_2016_az_1_1_2018.xlsx',
)


def load_cz_nuts4():
    data = get_remote_data(CZ_NUTS4_URL)

    # take last sheet
    sheet = data[list(data.keys())[-1]]

    nuts = {n.code: n for n in NUTS.objects.filter(code__startswith='CZ').iterator()}
    laus = {l.code: l for l in LAU.objects.filter(code__startswith='CZ').iterator()}
    nuts_created, nuts_updated, lau_created, lau_updated = 0, 0, 0, 0
    for row in sheet[2:]:
        lau_code, lau_name = row[0:2]
        nuts4_code, nuts4_name = row[7:9]
        lau_code = 'CZ%s' % lau_code
        if nuts4_code not in nuts:
            nuts[nuts4_code] = NUTS.objects.create(
                code=nuts4_code,
                name=nuts4_name,
                level=4,
            )
            nuts_created += 1
        elif nuts[nuts4_code].name != nuts4_name:
            nuts[nuts4_code].name = nuts4_name
            nuts[nuts4_code].save()
            nuts_updated += 1
        if lau_code not in laus:
            laus[lau_code] = LAU.objects.create(
                code=lau_code,
                name=lau_name,
                local_name=lau_name,
                nuts=nuts[nuts4_code],
            )
            lau_created += 1
        else:
            lau = laus[lau_code]
            if lau.name != lau_name or lau.local_name != lau_name or lau.nuts_id != nuts4_code:
                lau.name = lau_name
                lau.local_name = lau_name
                lau.nuts = nuts[nuts4_code]
                lau.save()
                lau_updated += 1
    logger.info(
        '%d NUTS created, %d updated, %d LAU created, %d updated',
        nuts_created, nuts_updated, lau_created, lau_updated,
    )
